/*
 *  drivers/mfd/ltc3577-pmic.c
 *
 *  Copyright (C) 2009 TomTom BV <http://www.tomtom.com/>
 *  Author: Andrzej Zukowski <andrzej.zukowski@tomtom.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/mfd/ltc3577/core.h>
#include <linux/mfd/ltc3577/pmic.h>

/*
 * Conversion macros
 */
#define LTC3577_GET_LED(burst, slew)		(burst | slew)

static	struct ltc3577           *ltc;

static ssize_t ltc3577_reg_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t len = 0;
	int count = 0;

	len += sprintf(buf + len, "%02X", ltc->reg_read(ltc));
	len += sprintf(buf + len, "\n");
	len += sprintf(buf + len, "\n");

	return len;
}

static ssize_t ltc3577_reg_store(struct device *dev, struct device_attribute *attr,  const char *buf, size_t count)
{
	int val;
	int reg;
	int ret;

	ret = sscanf(buf, "%x %x", &reg, &val);
	if (ret != 2) {
		printk(KERN_ERR "%s: wrong arguments\n", __func__);
		return 0;
	}

	ret = ltc->reg_write(ltc, reg, val);

	printk("%02X <-- %02X\n", reg, val);
	return count;
}

static DEVICE_ATTR(reg, 0644, ltc3577_reg_show, ltc3577_reg_store);

static void ltc3577_pmic_bucks(struct ltc3577 *ltc, uint8_t burst, uint8_t slew)
{
	ltc->reg_write(ltc, LTC3577_LED, LTC3577_GET_LED(burst, slew));
}

static int __devinit ltc3577_pmic_probe(struct platform_device *pdev)
{
	struct ltc3577_pmic_data *pdata;
	int ret;

	pdata = pdev->dev.platform_data;
	if (!pdata) {
		dev_err(&pdev->dev, "Failed to get platform data %s\n", pdev->name);
		ret = -ENODEV;
		goto err;
	}

	ltc = dev_to_ltc3577(&pdev->dev);
	if (!ltc) {
		dev_err(&pdev->dev, "Failed to get parent device %s\n", pdev->name);
		ret = -ENODEV;
		goto err;
	}

	if (!ltc->reg_read) {
		dev_err(&pdev->dev, "I2C not ready for receiving %s\n", pdev->name);
		ret = -ENODEV;
		goto err;
	}

	if (!ltc->reg_write) {
		dev_err(&pdev->dev, "I2C not ready for sending %s\n", pdev->name);
		ret = -ENODEV;
		goto err;
	}

	if ((device_create_file(&pdev->dev, &dev_attr_reg)) < 0) {
		dev_err(&pdev->dev, "failed to create reg file\n");
	}

	ltc3577_pmic_bucks(ltc, pdata->burst_mode, pdata->slew_rate);
	return 0;
err:
	return ret;
}

static int __devexit ltc3577_pmic_remove(struct platform_device *pdev)
{
	// anzu: add usb/gpio code here

	device_remove_file(&pdev->dev, &dev_attr_reg);
	return 0;
}

#ifdef CONFIG_PM
static int ltc3577_pmic_suspend(struct platform_device *pdev, pm_message_t state)
{
	// anzu: add usb code here
	return 0;
}

static int ltc3577_pmic_resume(struct platform_device *pdev)
{
	// anzu: add usb code here
	return 0;
}
#else
#define ltc3577_pmic_suspend	NULL
#define ltc3577_pmic_resume	NULL
#endif

static struct platform_driver ltc3577_pmic_driver = {
	.driver		= {
		.name	= "ltc3577-pmic",
		.owner	= THIS_MODULE,
	},
	.probe		= ltc3577_pmic_probe,
	.remove		= __devexit_p(ltc3577_pmic_remove),
	.suspend	= ltc3577_pmic_suspend,
	.resume		= ltc3577_pmic_resume,
};

static int __init ltc3577_pmic_init(void)
{
	return platform_driver_register(&ltc3577_pmic_driver);
}
module_init(ltc3577_pmic_init);

static void __exit ltc3577_pmic_exit(void)
{
	platform_driver_unregister(&ltc3577_pmic_driver);
}
module_exit(ltc3577_pmic_exit);

MODULE_DESCRIPTION("LTC3577 based PMIC Driver");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andrzej Zukowski <andrzej.zukowski@tomtom.com>");

